// mystbtt
// stb-truetype is an amazingly compact TrueType interpreter by Sean Barrett,
// part of his "stb" code collection. There are more good stuff in that collection
// but this was one that I really wanted, in order to produce high quality text
// with a compact, transparent code-only solution. Great work, Barrett!

// This is my interface to stb-truetype, the reusable parts, e.g. a library on top of the library.
// It is based on example code in stb-truetype but more in the style I needed.

// By Ingemar Ragnemalm. 2023 version with similar code isolation as SimpleFont/SimpleGUI,
// doing my best to be as non-intrusive as possible. NO additional libraries needed!

#define STB_TRUETYPE_IMPLEMENTATION  // force following include to generate implementation
#include "my-stbtt-lib.h"
#include <stdio.h>

static GLfloat vertices[] = {	-0.5f,-0.5f,0.0f,
						-0.5f,0.5f,0.0f,
						0.5f,-0.5f,0.0f,
						
						0.5f,-0.5f,0.0f,
						-0.5f,0.5f,0.0f,
						0.5f,0.5f,0.0f };

static GLfloat texcoord[] = {	0.0f, 1.0f,
						0.0f, 0.0f,
						1.0f, 1.0f,
						
						1.0f, 1.0f,
						0.0f, 0.0f,
						1.0f, 0.0f};

stbfont *my_stbtt_initfont(const char *fontName, float fontsize)
{
	stbfont *f = malloc(sizeof(stbfont));
	unsigned char temp_bitmap[512*512];
	unsigned char ttf_buffer[1<<20];
	
	FILE *ff = fopen(fontName, "rb");
	if (!ff)
	{
		printf("File not found: %s\n", fontName);
		return NULL;
	}
   fread(ttf_buffer, 1, 1<<20, ff); // MUST have error check here
   stbtt_BakeFontBitmap(ttf_buffer/*data*/, 0, fontsize, temp_bitmap,512,512, 32,96, f->cdata); // no guarantee this fits!
   fclose(ff);
   // can free ttf_buffer at this point
   glGenTextures(1, &f->ftex);
   glBindTexture(GL_TEXTURE_2D, f->ftex);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, 512,512, 0, GL_RED, GL_UNSIGNED_BYTE, temp_bitmap);
   // can free temp_bitmap at this point
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glGenerateMipmap(GL_TEXTURE_2D);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // Mipmapping pointless when magnifying
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	
	return f;
}

unsigned int vertexArrayObjID;
unsigned int vertexBufferObjID;
unsigned int texCoordBufferObjID;

// Compile a shader, return reference to it
// No error checking - the shaders should be stable
// Copied from SimpleFont/SimpleGUI
static GLuint sfCompileShaders(const char *vs, const char *fs)
{
	GLuint v,f,p;

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(v, 1, &vs, NULL);
	glShaderSource(f, 1, &fs, NULL);
	glCompileShader(v);
	glCompileShader(f);
	p = glCreateProgram();
	glAttachShader(p,v);
	glAttachShader(p,f);
	glLinkProgram(p);
	glUseProgram(p);
	return p;
}

// Inline shaders
static char frag[] =
"#version 150\n"
"out vec4 outColor;"
"in vec2 texCoord;"
"uniform sampler2D tex;"
"uniform vec3 color;"
"void main(void)"
"{"
"	vec4 pixel = texture(tex, texCoord);"
// Everything goes in red channel.
"	pixel = pixel.rrrr;"
"	outColor = pixel * vec4(color.r, color.g, color.b, 1.0);"
"}";


static char vert[]=
"#version 150\n"
"in  vec3 inPosition;"
"in vec2 inTexCoord;"
"out vec2 texCoord;"
"uniform mat4 m;"
"void main(void)"
"{"
"	texCoord = inTexCoord;"
"	gl_Position = m * vec4(inPosition, 1.0);"
"}";

static GLuint program;

typedef struct Saves
{
	GLint tex, activeTxtUnit;
	GLint saveprogram;
	char saveZ;
	char saveCull;
	char saveBlend;
	GLint savedSrcAlpha, savedDstAlpha;
};
struct Saves saves;

static void SaveState()
{
	// Get some current settings to restore them later
	saves.saveZ = glIsEnabled(GL_DEPTH_TEST);
	saves.saveCull = glIsEnabled(GL_CULL_FACE);
	saves.saveBlend = glIsEnabled(GL_BLEND);
	glGetIntegerv(GL_BLEND_SRC_ALPHA, &saves.savedSrcAlpha);
	glGetIntegerv(GL_BLEND_DST_ALPHA, &saves.savedDstAlpha);
	glGetIntegerv(GL_ACTIVE_TEXTURE, &saves.activeTxtUnit);
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &saves.tex);
	glGetIntegerv(GL_CURRENT_PROGRAM, &saves.saveprogram);

	glActiveTexture(GL_TEXTURE0);
}

static void RestoreState()
{
	glBindTexture(GL_TEXTURE_2D, saves.tex);
	glActiveTexture(saves.activeTxtUnit); // or should this be GL_TEXTURE0 + activeTxtUnit? Needs testing!
	glUseProgram(saves.saveprogram);
	if (saves.saveZ) glEnable(GL_DEPTH_TEST);
	if (saves.saveCull) glEnable(GL_CULL_FACE);
	if (!saves.saveBlend) glDisable(GL_BLEND);
	glBlendFunc(saves.savedSrcAlpha, saves.savedDstAlpha);
	glBindVertexArray(0); // Make sure the host program doesn't mess up our VAO
}

void my_stbtt_set_color(float r, float g, float b)
{
	GLint saveprogram;
	glGetIntegerv(GL_CURRENT_PROGRAM, &saveprogram);

	glUseProgram(program);
	glUniform3f(glGetUniformLocation(program, "color"), r, g, b);

	glUseProgram(saveprogram);
}

void my_stbtt_init()
{
	// Upload geometry to the GPU:
	
	program = sfCompileShaders(vert, frag);
	
	glUseProgram(program);
	// Allocate and activate Vertex Array Object
	glGenVertexArrays(1, &vertexArrayObjID);
	glBindVertexArray(vertexArrayObjID);
	// Allocate Vertex Buffer Objects
	glGenBuffers(1, &vertexBufferObjID);
	glGenBuffers(1, &texCoordBufferObjID);
	
	// VBO for vertex data
	glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObjID);
	glBufferData(GL_ARRAY_BUFFER, 18*sizeof(GLfloat), vertices, GL_STATIC_DRAW);
	glVertexAttribPointer(glGetAttribLocation(program, "inPosition"), 3, GL_FLOAT, GL_FALSE, 0, 0); 
	glEnableVertexAttribArray(glGetAttribLocation(program, "inPosition"));

	// VBO for texCoord data
	glBindBuffer(GL_ARRAY_BUFFER, texCoordBufferObjID);
	glBufferData(GL_ARRAY_BUFFER, 12*sizeof(GLfloat), texcoord, GL_STATIC_DRAW);
	glVertexAttribPointer(glGetAttribLocation(program, "inTexCoord"), 2, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(glGetAttribLocation(program, "inTexCoord"));

	glUniform1i(glGetUniformLocation(program, "tex"), 0); // Texture unit 0
	my_stbtt_set_color(1, 1, 1);
}

// In order to minimize dependencies, a few calls from VectorUtils:
	typedef struct internal_mat4
	{
		GLfloat m[16];
	} internal_mat4;
	
	static internal_mat4 internal_IdentityMatrix()
	{
		internal_mat4 m;
		int i;

		for (i = 0; i <= 15; i++)
			m.m[i] = 0;
		for (i = 0; i <= 3; i++)
			m.m[i * 5] = 1; // 0,5,10,15
		return m;
	}

	static internal_mat4 internal_T(GLfloat tx, GLfloat ty, GLfloat tz)
	{
		internal_mat4 m;
		m = internal_IdentityMatrix();
		m.m[3] = tx;
		m.m[7] = ty;
		m.m[11] = tz;
		return m;
	}

	static internal_mat4 internal_S(GLfloat sx, GLfloat sy, GLfloat sz)
	{
		internal_mat4 m;
		m = internal_IdentityMatrix();
		m.m[0] = sx;
		m.m[5] = sy;
		m.m[10] = sz;
		return m;
	}
	
	static internal_mat4 internal_Mult(internal_mat4 a, internal_mat4 b) // m = a * b
	{
		internal_mat4 m;

		int x, y;
		for (x = 0; x <= 3; x++)
			for (y = 0; y <= 3; y++)
					m.m[y*4 + x] =	a.m[y*4+0] * b.m[0*4+x] +
								a.m[y*4+1] * b.m[1*4+x] +
								a.m[y*4+2] * b.m[2*4+x] +
								a.m[y*4+3] * b.m[3*4+x];

		return m;
	}


float my_stbtt_print(float x, float y, const char *text, stbfont *f)
{
// Save state here
	SaveState();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

// Get size from OpenGL
	GLint a[4];
	glGetIntegerv(GL_VIEWPORT, a);
	int gRasterH = a[2]; // /gScale; Might be needed on M1 Macs
	int gRasterV = a[3]; // /gScale;

	internal_mat4 trans, rot, scale, m;
	scale = internal_S((float)1.0/gRasterH * 2, (float)-1.0/gRasterV * 2, 1);
	trans = internal_T(-1, 1, 0);
	m = internal_Mult(trans, scale);
	glUseProgram(program);
	glUniformMatrix4fv(glGetUniformLocation(program, "m"), 1, GL_TRUE, m.m);

//		printf("Start at %f %f printing '%s'\n", x, y, text);

   // uses orthographic projection with units = screen pixels, origin at top left
   glBindTexture(GL_TEXTURE_2D, f->ftex);

   stbtt_aligned_quad q;
   while (*text)
   {
      if (*text >= 32) //  && *text < 128)
      {
         stbtt_GetBakedQuad(f->cdata, 512,512, *text-32, &x,&y,&q,1);//1=opengl,0=old d3d

// Changes x and y?
// But then we can return that!
// But how about rotations...?
         
//		printf("%f %f %f %f\n", q.x0, q.y0, q.x1, q.y1);
         
// One drawing call for every character isn't the most optimal case I can imagine.
// It should be possible to handle this with instancing. Future work!

         vertices[0] = q.x0;
         vertices[1] = q.y0;
         vertices[3] = q.x0;
         vertices[4] = q.y1;
         vertices[6] = q.x1;
         vertices[7] = q.y0;
         
         vertices[9] = q.x1;
         vertices[10] = q.y0;
         vertices[12] = q.x0;
         vertices[13] = q.y1;
         vertices[15] = q.x1;
         vertices[16] = q.y1;
         
         texcoord[0] = q.s0;
         texcoord[1] = q.t0;
         texcoord[2] = q.s0;
         texcoord[3] = q.t1;
         texcoord[4] = q.s1;
         texcoord[5] = q.t0;
         
         texcoord[6] = q.s1;
         texcoord[7] = q.t0;
         texcoord[8] = q.s0;
         texcoord[9] = q.t1;
         texcoord[10] = q.s1;
         texcoord[11] = q.t1;

			glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObjID);
			glBufferData(GL_ARRAY_BUFFER, 18*sizeof(GLfloat), vertices, GL_STATIC_DRAW);
			glBindBuffer(GL_ARRAY_BUFFER, texCoordBufferObjID);
			glBufferData(GL_ARRAY_BUFFER, 12*sizeof(GLfloat), texcoord, GL_STATIC_DRAW);
			
			glBindVertexArray(vertexArrayObjID);	// Select VAO
			glDrawArrays(GL_TRIANGLES, 0, 6);	// draw object
      }
      ++text;
   }

   stbtt_GetBakedQuad(f->cdata, 512,512, 0, &x,&y,&q,1); // Trick to get the proper next position - I think. Otherwise it will ignore ending spaces.

// Restore state here
	RestoreState();
	return q.x1; // Try to tell how far it went.
}
